# pylint: disable=unused-argument,missing-function-docstring
# pylint: disable=redefined-outer-name, unused-variable
# pylint: disable=redefined-builtin, missing-class-docstring,
import argparse
import inspect
import logging
import os
import sys
import uuid
from enum import Enum
from typing import List

import uvicorn
from fastapi import Depends, FastAPI, HTTPException, Query, Request, Security
from fastapi.responses import Response
from fastapi.security.api_key import APIKeyHeader, APIKeyQuery
from starlette import status
from suzieq.shared.exceptions import UserQueryError
from suzieq.shared.utils import (DATA_FORMATS, get_log_params,
                                 get_sq_install_dir, load_sq_config,
                                 print_version, sq_get_config_file)
from suzieq.sqobjects import get_sqobject

API_KEY_NAME = 'access_token'

api_key_query = APIKeyQuery(name=API_KEY_NAME, auto_error=False)
api_key_header = APIKeyHeader(name=API_KEY_NAME, auto_error=False)


def check_config_file():
    if not getattr(app, 'cfg_file', None):
        print('missing config file')
        sys.exit(1)


# Changing the default URLs to help with reverse proxy stuff as described
# in issue #381 (https://github.com/netenglabs/suzieq/issues/381)
app = FastAPI(on_startup=[check_config_file],
              openapi_url="/api/openapi.json",
              docs_url="/api/docs",
              redoc_url="/api/redoc")


def app_init(cfg_file):
    '''This is the actual API initilaizer'''
    # pylint: disable=global-variable-not-assigned

    global app

    app.cfg_file = cfg_file

    return app


def get_configured_api_key():
    cfg = load_sq_config(config_file=app.cfg_file)
    try:
        api_key = cfg['rest']['API_KEY']
    except KeyError:
        print('missing API_KEY in config file')
        sys.exit(1)

    return api_key


async def get_api_key(api_key_query: str = Security(api_key_query),
                      api_key_header: str = Security(api_key_header)):

    api_key = get_configured_api_key()
    if api_key_query == api_key:
        return api_key_query
    elif api_key_header == api_key:
        return api_key_header
    else:

        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid API Key",
        )


def get_cert_files(cfg):
    sqdir = get_sq_install_dir()
    ssl_certfile = cfg.get('rest', {}) \
                      .get('rest-certfile', f'{sqdir}/config/etc/cert.pem')

    ssl_keyfile = cfg.get('rest', {}) \
                     .get('rest-keyfile', f'{sqdir}/config/etc/key.pem')

    if not os.path.isfile(ssl_certfile):
        print(f"ERROR: Missing certificate file: {ssl_certfile}")
        sys.exit(1)

    if not os.path.isfile(ssl_keyfile):
        print(f"ERROR: Missing certificate file: {ssl_keyfile}")
        sys.exit(1)

    return ssl_keyfile,  ssl_certfile


def get_log_config_level(cfg):

    logfile, loglevel, logsize, log_stdout = get_log_params(
        'rest', cfg, '/tmp/sq-rest-server.log')

    log_config = uvicorn.config.LOGGING_CONFIG
    if logfile and not log_stdout:
        log_config['handlers']['access']['filename'] = logfile
        log_config['handlers']['access']['class'] = \
            'logging.handlers.RotatingFileHandler'
        log_config['handlers']['access']['maxBytes'] = logsize
        log_config['handlers']['access']['backupCount'] = 2

        log_config['handlers']['default']['class'] = \
            'logging.handlers.RotatingFileHandler'
        log_config['handlers']['default']['maxBytes'] = logsize
        log_config['handlers']['default']['backupCount'] = 2
        log_config['handlers']['default']['filename'] = logfile

        if 'stream' in log_config['handlers']['default']:
            del log_config['handlers']['default']['stream']
            del log_config['handlers']['access']['stream']

    return log_config, loglevel


def rest_main(*args) -> None:
    """The main function for the REST server

    Args:
        config_file (str): The suzieq config file
        no_https (bool): If true, disable https
    """

    if not args:
        args = sys.argv

    parser = argparse.ArgumentParser(args)
    parser.add_argument(
        "-c",
        "--config",
        type=str, help="alternate config file",
        default=None
    )
    parser.add_argument(
        "--no-https",
        help="Turn off HTTPS",
        default=False, action='store_true',
    )
    parser.add_argument(
        "--version",
        "-V",
        help="print Suzieq version",
        default=False, action='store_true',
    )
    userargs = parser.parse_args()

    if userargs.version:
        print_version()
        sys.exit(0)

    config_file = sq_get_config_file(userargs.config)
    app = app_init(config_file)
    cfg = load_sq_config(config_file=config_file)
    try:
        _ = cfg['rest']['API_KEY']
    except KeyError:
        print('missing API_KEY in config file')
        sys.exit(1)

    logcfg, loglevel = get_log_config_level(cfg)

    no_https = cfg.get('rest', {}).get('no-https', False) or userargs.no_https

    srvr_addr = cfg.get('rest', {}).get('address', '127.0.0.1')
    srvr_port = cfg.get('rest', {}).get('port', 8000)

    if no_https:
        uvicorn.run(app, host=srvr_addr, port=srvr_port,
                    )
    else:
        ssl_keyfile, ssl_certfile = get_cert_files(cfg)
        uvicorn.run(app, host=srvr_addr, port=srvr_port,
                    ssl_keyfile=ssl_keyfile,
                    ssl_certfile=ssl_certfile)


class CommonVerbs(str, Enum):
    show = "show"
    summarize = "summarize"
    unique = "unique"
    top = "top"


class CommonExtraVerbs(str, Enum):
    aver = "assert"
    show = "show"
    summarize = "summarize"
    unique = "unique"
    top = "top"


class RouteVerbs(str, Enum):
    show = "show"
    summarize = "summarize"
    unique = "unique"
    lpm = "lpm"
    top = "top"


class NetworkVerbs(str, Enum):
    find = "find"


class DeviceStatus(str, Enum):
    alive = "alive"
    dead = "dead"
    neverpoll = "neverpoll"
    notalive = "!alive"
    notdead = "!dead"
    notneverpoll = "!neverpoll"


class BgpStateValues(str, Enum):
    ESTABLISHED = "Established"
    NOTESTD = "NotEstd"
    DYNAMIC = "dynamic"
    NOTESTABLISHED = "!Established"
    NOTNOTESTD = "!NotEstd"
    NOTDYNAMIC = "!dynamic"


class IfStateValues(str, Enum):
    UP = "up"
    DOWN = "down"
    ERRDISABLED = "errDisabled"
    NOTCONNECTED = "notConnected"
    NOTUP = "!up"
    NOTDOWN = "!down"
    NOTERRDISABLED = "!errDisabled"
    CONNECTED = "!notConnected"


class OspfStateValues(str, Enum):
    FULL = "full"
    PASSIVE = "passive"
    OTHER = "other"
    NOTFULL = "!full"
    NOTPASSIVE = "!passive"
    NOTOTHER = "!other"


class ViewValues(str, Enum):
    latest = "latest"
    all = "all"
    changes = "changes"


class AssertResultValue(str, Enum):
    PASS = "pass"
    FAIL = "fail"
    ALL = "all"


class SqPollerStatus(str, Enum):
    PASS = "pass"
    FAIL = "fail"
    ALL = "all"


class InventoryStatusValues(str, Enum):
    PRESENT = "present"
    ABSENT = "absent"


class TruthasStrings(str, Enum):
    yes = "True"
    no = "False"

# The logic in the code below is that you have a common function to
# split the common arguments across all the functions, and split the
# object-specific arguments in the object function itself.


@app.get("/api/v1/{rest_of_path:path}", deprecated=True)
async def deprecated_function(request: Request, rest_of_path: str):
    return([{'error': 'v1 is deprecated, use API version v2'}])


@app.get("/api/v2/address/{verb}")
async def query_address(verb: CommonVerbs, request: Request,
                        token: str = Depends(get_api_key),
                        format: str = None,
                        hostname: List[str] = Query(None),
                        start_time: str = "", end_time: str = "",
                        view: ViewValues = "latest",
                        namespace: List[str] = Query(None),
                        columns: List[str] = Query(default=["default"]),
                        address: List[str] = Query(None),
                        type: List[str] = Query(None),
                        ifname: List[str] = Query(None),
                        prefix: List[str] = Query(None),
                        ipvers: str = None, what: str = None,
                        vrf: List[str] = Query(None), query_str: str = None,
                        count: str = None, reverse: str = None,
                        ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/arpnd/{verb}")
async def query_arpnd(verb: CommonVerbs, request: Request,
                      token: str = Depends(get_api_key),
                      format: str = None,
                      hostname: List[str] = Query(None),
                      start_time: str = "", end_time: str = "",
                      view: ViewValues = "latest",
                      namespace: List[str] = Query(None),
                      columns: List[str] = Query(default=["default"]),
                      ipAddress: List[str] = Query(None),
                      macaddr: List[str] = Query(None),
                      prefix: List[str] = Query(None),
                      oif: List[str] = Query(None),
                      query_str: str = None, what: str = None,
                      count: str = None, reverse: str = None,
                      ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/bgp/{verb}")
async def query_bgp(verb: CommonExtraVerbs, request: Request,
                    token: str = Depends(get_api_key),
                    format: str = None,
                    hostname: List[str] = Query(None),
                    start_time: str = "", end_time: str = "",
                    view: ViewValues = "latest",
                    namespace: List[str] = Query(None),
                    columns: List[str] = Query(default=["default"]),
                    peer: List[str] = Query(None),
                    state: BgpStateValues = Query(None),
                    vrf: List[str] = Query(None),
                    asn: List[str] = Query(None),
                    result: AssertResultValue = Query(None),
                    afiSafi: str = Query(None),
                    query_str: str = None, what: str = None,
                    count: str = None, reverse: str = None,
                    ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/device/{verb}")
async def query_device(verb: CommonVerbs, request: Request,
                       token: str = Depends(get_api_key),
                       format: str = None,
                       hostname: List[str] = Query(None),
                       start_time: str = "", end_time: str = "",
                       view: ViewValues = "latest",
                       namespace: List[str] = Query(None),
                       columns: List[str] = Query(default=["default"]),
                       query_str: str = None,
                       os: List[str] = Query(None),
                       vendor: List[str] = Query(None),
                       model: List[str] = Query(None),
                       version: List[str] = Query(None),
                       what: str = None,
                       status: List[DeviceStatus] = Query(None),
                       count: str = None, reverse: str = None,
                       ):
    function_name = inspect.currentframe().f_code.co_name
    if status:
        status = [x.value for x in status]  # convert enum to string
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/devconfig/{verb}")
async def query_devconfig(verb: CommonVerbs, request: Request,
                          token: str = Depends(get_api_key),
                          format: str = None,
                          hostname: List[str] = Query(None),
                          start_time: str = "", end_time: str = "",
                          view: ViewValues = "latest",
                          namespace: List[str] = Query(None),
                          columns: List[str] = Query(default=["default"]),
                          query_str: str = None,
                          what: str = None,
                          count: str = None, reverse: str = None,
                          ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/evpnVni/{verb}")
async def query_evpnVni(verb: CommonExtraVerbs, request: Request,
                        token: str = Depends(get_api_key),
                        format: str = None,
                        hostname: List[str] = Query(None),
                        start_time: str = "", end_time: str = "",
                        view: ViewValues = "latest",
                        namespace: List[str] = Query(None),
                        columns: List[str] = Query(default=["default"]),
                        vni: List[str] = Query(None),
                        priVtepIp: List[str] = Query(None),
                        result: AssertResultValue = None,
                        query_str: str = None, what: str = None,
                        count: str = None, reverse: str = None,
                        ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/fs/{verb}")
async def query_fs(verb: CommonVerbs, request: Request,
                   token: str = Depends(get_api_key),
                   format: str = None,
                   hostname: List[str] = Query(None),
                   start_time: str = "", end_time: str = "",
                   view: ViewValues = "latest",
                   namespace: List[str] = Query(None),
                   columns: List[str] = Query(default=["default"]),
                   mountPoint: List[str] = Query(None), what: str = None,
                   usedPercent: str = None, query_str: str = None,
                   count: str = None, reverse: str = None,
                   ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/interface/{verb}")
async def query_interface(verb: CommonExtraVerbs, request: Request,
                          token: str = Depends(get_api_key),
                          format: str = None,
                          hostname: List[str] = Query(None),
                          start_time: str = "", end_time: str = "",
                          view: ViewValues = "latest",
                          namespace: List[str] = Query(None),
                          columns: List[str] = Query(default=["default"]),
                          ifname: List[str] = Query(None),
                          state: IfStateValues = Query(None),
                          type: List[str] = Query(None),
                          what: str = None, vrf: List[str] = Query(None),
                          master: List[str] = Query(None),
                          mtu: List[str] = Query(None),
                          ifindex: List[str] = Query(None),
                          matchval: int = Query(None, alias="value"),
                          result: AssertResultValue = Query(None),
                          ignore_missing_peer: bool = Query(False),
                          vlan: List[str] = Query(None),
                          portmode: List[str] = Query(None),
                          query_str: str = None,
                          count: str = None, reverse: str = None,
                          ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/inventory/{verb}")
async def query_inventory(verb: CommonVerbs, request: Request,
                          token: str = Depends(get_api_key),
                          format: str = None,
                          hostname: List[str] = Query(None),
                          start_time: str = "", end_time: str = "",
                          view: ViewValues = "latest",
                          namespace: List[str] = Query(None),
                          columns: List[str] = Query(default=["default"]),
                          query_str: str = None,
                          type: List[str] = Query(None),
                          serial: List[str] = Query(None),
                          model: List[str] = Query(None),
                          vendor: List[str] = Query(None), what: str = None,
                          status: InventoryStatusValues = Query(None),
                          count: str = None, reverse: str = None,
                          ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/lldp/{verb}")
async def query_lldp(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     peerMacaddr: List[str] = Query(None),
                     peerHostname: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     ifname: List[str] = Query(None),
                     use_bond: TruthasStrings = Query(None),
                     query_str: str = None, what: str = None,
                     count: str = None, reverse: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/mac/{verb}")
async def query_mac(verb: CommonVerbs, request: Request,
                    token: str = Depends(get_api_key),
                    format: str = None,
                    hostname: List[str] = Query(None),
                    start_time: str = "", end_time: str = "",
                    view: ViewValues = "latest",
                    namespace: List[str] = Query(None),
                    columns: List[str] = Query(default=["default"]),
                    bd: str = None,
                    local: str = None,
                    macaddr: List[str] = Query(None),
                    remoteVtepIp: List[str] = Query(None),
                    vlan: List[str] = Query(None),
                    query_str: str = None, what: str = None,
                    moveCount: str = None, count: str = None,
                    reverse: str = None,
                    ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/mlag/{verb}")
async def query_mlag(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     query_str: str = None, what: str = None,
                     count: str = None, reverse: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/network/show", deprecated=True)
async def query_network_depr_show(request: Request,
                                  token: str = Depends(get_api_key),
                                  format: str = None,
                                  columns: List[str] = Query(
                                      default=["default"]),
                                  namespace: List[str] = Query(None),
                                  hostname: List[str] = Query(None),
                                  start_time: str = "", end_time: str = "",
                                  version: str = "",
                                  view: ViewValues = "latest",
                                  model: List[str] = Query(None),
                                  vendor: List[str] = Query(None),
                                  os: List[str] = Query(None),
                                  query_str: str = None, what: str = None,
                                  count: str = None, reverse: str = None,
                                  ):
    function_name = 'query_namespace'
    verb = 'show'
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/network/summarize", deprecated=True)
async def query_network_depr_summarize(request: Request,
                                       token: str = Depends(get_api_key),
                                       format: str = None,
                                       columns: List[str] = Query(
                                           default=["default"]),
                                       namespace: List[str] = Query(None),
                                       hostname: List[str] = Query(None),
                                       start_time: str = "",
                                       end_time: str = "",
                                       version: str = "",
                                       view: ViewValues = "latest",
                                       model: List[str] = Query(None),
                                       vendor: List[str] = Query(None),
                                       os: List[str] = Query(None),
                                       query_str: str = None,
                                       what: str = None,
                                       count: str = None,
                                       reverse: str = None,
                                       ):
    function_name = 'query_namespace'
    verb = 'summarize'
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/network/unique", deprecated=True)
async def query_network_depr_unique(request: Request,
                                    token: str = Depends(get_api_key),
                                    format: str = None,
                                    columns: List[str] = Query(
                                        default=["default"]),
                                    namespace: List[str] = Query(None),
                                    hostname: List[str] = Query(None),
                                    start_time: str = "",
                                    end_time: str = "",
                                    version: str = "",
                                    view: ViewValues = "latest",
                                    model: List[str] = Query(None),
                                    vendor: List[str] = Query(None),
                                    os: List[str] = Query(None),
                                    query_str: str = None,
                                    what: str = None,
                                    count: str = None, reverse: str = None,
                                    ):
    function_name = 'query_namespace'
    verb = 'unique'
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/network/top", deprecated=True)
async def query_network_depr_top(request: Request,
                                 token: str = Depends(get_api_key),
                                 format: str = None,
                                 columns: List[str] = Query(
                                     default=["default"]),
                                 namespace: List[str] = Query(None),
                                 hostname: List[str] = Query(None),
                                 start_time: str = "", end_time: str = "",
                                 version: str = "",
                                 view: ViewValues = "latest",
                                 model: List[str] = Query(None),
                                 vendor: List[str] = Query(None),
                                 os: List[str] = Query(None),
                                 query_str: str = None, what: str = None,
                                 count: str = None, reverse: str = None,
                                 ):
    function_name = 'query_namespace'
    verb = 'top'
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/network/{verb}")
async def query_network(verb: NetworkVerbs, request: Request,
                        token: str = Depends(get_api_key),
                        format: str = None,
                        columns: List[str] = Query(default=["default"]),
                        namespace: List[str] = Query(None),
                        hostname: List[str] = Query(None),
                        start_time: str = "", end_time: str = "",
                        view: ViewValues = "latest",
                        address: List[str] = Query(None),
                        vlan: str = '', vrf: str = '',
                        query_str: str = None,
                        ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/namespace/{verb}")
async def query_namespace(verb: CommonVerbs, request: Request,
                          token: str = Depends(get_api_key),
                          format: str = None,
                          columns: List[str] = Query(default=["default"]),
                          namespace: List[str] = Query(None),
                          hostname: List[str] = Query(None),
                          start_time: str = "", end_time: str = "",
                          version: str = "",
                          view: ViewValues = "latest",
                          model: List[str] = Query(None),
                          vendor: List[str] = Query(None),
                          os: List[str] = Query(None),
                          query_str: str = None, what: str = None,
                          count: str = None, reverse: str = None,
                          ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/ospf/{verb}")
async def query_ospf(verb: CommonExtraVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     ifname: List[str] = Query(None),
                     state: OspfStateValues = Query(None),
                     area: List[str] = Query(None),
                     vrf: List[str] = Query(None),
                     result: AssertResultValue = None,
                     query_str: str = None, what: str = None,
                     count: str = None, reverse: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/path/{verb}")
async def query_path(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     vrf: str = Query(None),
                     dest: str = Query(None),
                     src: str = Query(None),
                     query_str: str = None, what: str = None,
                     count: str = None, reverse: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/route/{verb}")
async def query_route(verb: RouteVerbs, request: Request,
                      token: str = Depends(get_api_key),
                      format: str = None,
                      hostname: List[str] = Query(None),
                      start_time: str = "", end_time: str = "",
                      view: ViewValues = "latest",
                      namespace: List[str] = Query(None),
                      columns: List[str] = Query(default=["default"]),
                      prefix: List[str] = Query(None),
                      vrf: List[str] = Query(None),
                      protocol: List[str] = Query(None),
                      prefixlen: str = None, ipvers: str = None,
                      add_filter: str = None, address: str = None,
                      query_str: str = None, what: str = None,
                      count: str = None, reverse: str = None,
                      ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/sqPoller/{verb}")
async def query_sqPoller(verb: CommonVerbs, request: Request,
                         token: str = Depends(get_api_key),
                         format: str = None,
                         hostname: List[str] = Query(None),
                         start_time: str = "", end_time: str = "",
                         view: ViewValues = "latest",
                         namespace: List[str] = Query(None),
                         columns: List[str] = Query(default=["default"]),
                         service: str = None,
                         status: SqPollerStatus = Query(None),
                         query_str: str = None, what: str = None,
                         pollExcdPeriodCount: str = None,
                         count: str = None, reverse: str = None,
                         ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/topology/{verb}")
async def query_topology(verb: CommonVerbs, request: Request,
                         token: str = Depends(get_api_key),
                         format: str = None,
                         hostname: List[str] = Query(None),
                         start_time: str = "", end_time: str = "",
                         view: ViewValues = "latest",
                         namespace: List[str] = Query(None),
                         columns: List[str] = Query(default=["default"]),
                         polled: str = None,
                         via: List[str] = Query(None),
                         ifname: List[str] = Query(None),
                         peerHostname: List[str] = Query(None),
                         asn: List[str] = Query(None),
                         area: List[str] = Query(None),
                         vrf: List[str] = Query(None),
                         afiSafi: str = Query(None),
                         query_str: str = None, what: str = None,
                         count: str = None, reverse: str = None,
                         ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/table/describe")
async def query_table_describe(
        request: Request,
        token: str = Depends(get_api_key),
        format: str = None,
        table: str = None,
):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, 'describe', request, locals())


@app.get("/api/v2/table/{verb}")
async def query_table(
        verb: CommonVerbs, request: Request,
        token: str = Depends(get_api_key),
        format: str = None,
        hostname: List[str] = Query(None),
        start_time: str = "", end_time: str = "",
        view: ViewValues = "latest", namespace: List[str] = Query(None),
        columns: List[str] = Query(default=["default"]),
        query_str: str = None, table: str = None,
        what: str = None,
        count: str = None, reverse: str = None,
):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/vlan/{verb}")
async def query_vlan(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     vlan: List[str] = Query(None),
                     state: str = None,
                     vlanName: List[str] = Query(None),
                     query_str: str = None, what: str = None,
                     count: str = None, reverse: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


def read_shared(function_name, verb, request, local_variables=None):
    """all the shared code for each of thse read functions"""

    command = function_name.split('_')[1]  # assumes fn name is query_<command>
    command_args, verb_args = create_filters(function_name, command, request,
                                             local_variables)

    verb = cleanup_verb(verb)

    columns = local_variables.get('columns', None)
    format = local_variables.get('format', None)
    if not format:
        format = 'json'
    if format not in DATA_FORMATS:
        return_error(405, f"Unsupported output format '{format}'")

    ret, svc_inst = run_command_verb(
        command, verb, command_args, verb_args, columns, format)

    return ret


def create_filters(function_name, command, request, local_vars):
    command_args = {}
    verb_args = {}
    remove_args = ['verb', 'token', 'format', 'request', 'access_token']
    all_cmd_args = ['namespace', 'hostname',
                    'start_time', 'end_time', 'view', 'columns', 'format']
    both_verb_and_command = ['namespace', 'hostname', 'columns']

    query_ks = request.query_params
    for arg in query_ks.keys():
        if arg in remove_args:
            continue
        if arg in all_cmd_args:
            if query_ks.get(arg) is not None:
                command_args[arg] = local_vars.get(arg, None)
                if arg in both_verb_and_command:
                    verb_args[arg] = command_args[arg]
        else:
            if query_ks.get(arg) is not None:
                verb_args[arg] = local_vars.get(arg, None)

    return command_args, verb_args


def cleanup_verb(verb):
    if verb == 'show':
        verb = 'get'
    if verb == 'assert':
        verb = 'aver'
    return verb


def create_command_args(hostname='', start_time='', end_time='', view='latest',
                        namespace='', columns='default'):
    command_args = {'hostname': hostname,
                    'start_time': start_time,
                    'end_time': end_time,
                    'view': view,
                    'namespace': namespace,
                    'columns': columns}
    return command_args


def get_svc(command):
    """based on the command, find the module and service that the command is in
    return the service
    """
    command_name = command

    svc = get_sqobject(command_name)
    return svc


def run_command_verb(command, verb, command_args, verb_args,
                     columns=None, format=None):
    """
    Runs the command and verb with the command_args and verb_args

    HTTP Return Codes
        404 -- Missing command or argument (including missing valid path)
        405 -- Missing or incorrect query parameters
        422 -- FastAPI validation errors
        500 -- Exceptions
    """

    if columns is None:
        columns = ['default']

    svc = get_svc(command)
    try:
        svc_inst = svc(**command_args,
                       config_file=app.cfg_file,
                       engine_name="pandas")
        df = getattr(svc_inst, verb)(**verb_args)

    except AttributeError as err:
        return_error(
            404, (f"{err}"))

    except NotImplementedError as err:
        return_error(404, f"{verb} not supported for {command}: {err}")

    except TypeError as err:
        return_error(405, f"bad keyword/filter for {command} {verb}: {err}")

    except ValueError as err:
        return_error(405, f"bad keyword/filter for {command} {verb}: {err}")

    except UserQueryError as err:
        return_error(500, f'UserQueryError: {err}')
    except Exception as err:
        return_error(
            500,
            f"{err}")

    if df.columns.to_list() == ['error']:
        return_error(
            405, f"bad keyword/filter for {command} {verb}: {df['error'][0]}")

    res_content = None
    media_type = None
    if format == 'markdown':
        # have to return a Reponse so that it won't turn the markdown into JSON
        res_content = df.to_markdown()
        media_type = 'text/plain'
    elif format == 'csv':
        res_content = df.to_csv()
        media_type = 'text/csv'
    elif format == 'text':
        res_content = df.to_string()
        media_type = 'text/plain'
    elif format == 'json':
        if verb == 'summarize':
            json_orient = 'columns'
        else:
            json_orient = 'records'
        media_type = 'application/json'
        res_content = df.to_json(orient=json_orient)
    return Response(content=res_content, media_type=media_type), svc_inst


def return_error(code: int, msg: str):
    u = uuid.uuid1()
    msg = f"{msg} id={u}"
    logger = logging.getLogger('uvicorn')
    logger.info(msg)
    raise HTTPException(status_code=code, detail=msg)


@ app.get("/api/v2/{command}", include_in_schema=False)
def missing_verb(command):
    return_error(
        404, f'{command} command missing a verb. for example '
        f'/api/v2/{command}/show')


@ app.get("/", include_in_schema=False)
def bad_path():
    return_error(
        404,
        "bad path. Try something like '/api/v2/device/show' or '/api/docs'")



from abc import abstractmethod
from suzieq.shared.sq_plugin import SqPlugin


class InventoryAsyncPlugin(SqPlugin):
    """Plugins which inherit this class will have methods 'run'

    Once the controller check that the object inherit this class, it launches
    a new task executing the run method.
    """

    async def run(self):
        """Background task to launch in order to execute the plugin"""
        try:
            await self._execute()
        finally:
            await self._stop()

    @abstractmethod
    async def _execute(self):
        """Launch the backuground task
        """

    async def _stop(self):
        """Actions to execute before terminating the task
        """
        return



'''
Parsing Configuration utilities
'''
from typing import Dict
from ciscoconfparse import CiscoConfParse


def get_access_port_interfaces(conf: CiscoConfParse,
                               nos: str) -> Dict[str, int]:
    '''For various NOS return the list of access port interfaces.

    This module uses the CiscoConfParse to extract the interface names.
    '''
    if 'junos' in nos:
        ifmap = _get_swport_interfaces_junos(conf, what='access')
    elif 'cumulus' in nos:
        ifmap = _get_swport_interfaces_cls(conf, what='access')
    else:
        ifmap = _get_swport_interfaces_iosy(conf, what='access')

    return ifmap


def get_trunk_port_interfaces(conf: CiscoConfParse,
                              nos: str) -> Dict[str, int]:
    '''For various NOS return the list of access port interfaces.

    This module uses the CiscoConfParse to extract the interface names.
    '''
    if 'junos' in nos:
        ifmap = _get_swport_interfaces_junos(conf, what='trunk')
    elif 'cumulus' in nos:
        ifmap = _get_swport_interfaces_cls(conf, what='trunk')
    else:
        ifmap = _get_swport_interfaces_iosy(conf, what='trunk')

    return ifmap


# Do not invoke these functions directly, the implementation can change
def _get_swport_interfaces_junos(conf: CiscoConfParse,
                                 what: str) -> Dict[str, int]:
    '''Return interfaces that match the requested info for Junos.

    This involves parsing the Junos config looking for the relevant info.
    what can be access or trunk. For access ports, we return a Dict[str, int]
    of interfaces that are access ports with the value that is the access vlan.
    '''

    pm_dict = {}
    if conf.syntax != 'junos':
        raise ValueError('Invalid config passed for Junos')

    if what == 'access':
        block = conf.find_objects_w_all_children('', ['interface-mode access'],
                                                 recurse=True)
        # The format returned is 3 lines for each interface with the first
        # line being the interface name
        for index in range(1, len(block), 3):
            ifname = block[index].text.strip()
            vlan_ln = [line for line in block[index].lineage
                       if 'members' in line.text]
            if vlan_ln:
                vlan = vlan_ln[0].text.split('members')[1].strip()
                if vlan.isnumeric():
                    pm_dict[ifname] = int(vlan)
                else:
                    vlan = 0

    return pm_dict


def _get_swport_interfaces_cls(conf: CiscoConfParse,
                               what: str) -> Dict[str, int]:
    '''Return interfaces that match the requested info for Cumulus.

    This involves parsing the ifupdown2 config looking for the relevant info.
    what can be access or trunk. For access ports, we return a dict that
    contains the access vlan for the interface.
    '''
    pm_dict = {}
    if what == 'access':
        for intf in conf.find_objects_w_all_children(
                '^iface ', ['bridge-access']):
            ifname = intf.text.split('iface')[1].strip()
            acc_vlan = intf.re_match_iter_typed(r'^.*bridge-access\s+(\d+)')
            if acc_vlan.isnumeric():
                pm_dict[ifname] = int(acc_vlan)
            else:
                pm_dict[ifname] = 0
    if what == "trunk":
        acc_ports = []
        for intf in conf.find_objects_w_all_children(
                '^iface ', ['bridge-access']):
            acc_ports.append(intf.text.split('iface')[1].strip())

        bridge_port_lines = conf.find_lines('bridge-ports')
        for line in bridge_port_lines:
            if line.strip().startswith('#'):
                continue
            # This line looks like this:
            # bridge-ports bond01 bond02 peerlink vni13 vni24 vxlan4001
            # Ignore the keyword, and trunk ports are those that are not
            # access ports
            brports = line.split()[1:]
            pm_dict = {port: '1' for port in brports
                       if port not in acc_ports}

        # Now we need to find the PVID for the ports
        for intf in conf.find_objects_w_child('iface ', 'bridge-pvid'):
            port = intf.text.split('iface')[1].strip()
            if port not in pm_dict:
                vlan = intf.re_match_iter_typed(r'bridge-pvid\s+(\d+)')
            else:
                vlan = intf.re_match_iter_typed(r'bridge-pvid\s+(\d+)')
            if vlan.isnumeric():
                pm_dict['port'] = int(vlan)
            else:
                pm_dict['port'] = 0
    return pm_dict


def _get_swport_interfaces_iosy(conf: CiscoConfParse,
                                what: str) -> Dict[str, int]:
    '''Return interfaces that match the requested info for IOSy NOS.

    This involves parsing the config looking for the relevant info. This is
    for devices which use an IOS-like config such as EOS, NXOS, IOSXR etc.
    what can be access or trunk. For access ports, we return a dict that
    contains the access vlan for the interface.
    '''

    pm_dict = {}
    if what == 'access':
        for intf in conf.find_objects_w_child('^interface ', '.*access'):
            ifname = intf.text.split('interface')[1].strip()
            acc_vlan = intf.re_match_iter_typed(r'^.*vlan\s+(\d+)')
            if acc_vlan.isnumeric():
                pm_dict[ifname] = int(acc_vlan)
            else:
                pm_dict[ifname] = 0

    if what == 'trunk':
        for intf in conf.find_objects_w_child('^interface', '.*trunk'):
            ifname = intf.text.split('interface')[1].strip()
            nvlan = intf.re_match_iter_typed(r'.*native vlan\s+(\d+)',
                                             default='1')
            if nvlan.isnumeric():
                pm_dict[ifname] = int(nvlan)
            else:
                pm_dict[ifname] = 0

    return pm_dict



import typing
import os
from nubia import command, context
from nubia.internal.commands.help import HelpCommand
from nubia.internal.cmdbase import Command
from prompt_toolkit.completion import Completion
from termcolor import cprint, colored

from suzieq.cli.nubia_patch import argument
from suzieq.shared.utils import SUPPORTED_ENGINES, print_version


@command("set")
@argument("namespace", description="namespace to qualify selection")
@argument("hostname", description="Name of host to qualify selection")
@argument(
    "start_time",
    description="Start of time window in YYYY-MM-dd HH:mm:SS format"
)
@argument("end_time",
          description="End of time window in YYYY-MM-dd HH:mm:SS format")
@argument("pager", description="Enable pagination prompt on longer outputs",
          choices=['on', 'off'])
@argument('col_width', description='Max Width of each column in table display')
@argument('max_rows',
          description='Max rows to display before ellipsis is shown')
@argument('all_columns',
          description='Display all columns wrapping around if necessary',
          choices=['yes', 'no'])
@argument(
    "engine",
    choices=SUPPORTED_ENGINES,
    description="What backend analyzer engine to use",
)
@argument(
    "debug", choices=['True'], description="Enable debug mode",
)
@argument(
    "datadir",
    description="Set the data directory for the command"
)
@argument(
    'rest_server_ip', description='IP address of the REST server'
)
@argument(
    'rest_server_port', description='Port of the REST server'
)
@argument(
    'rest_api_key', description='API key for the REST server'
)
@argument(
    'rest_use_https', description='Use HTTPS for the REST server',
    choices=['True', 'False']
)
def set_ctxt(
        pager: str = "",
        hostname: typing.List[str] = None,
        start_time: str = "",
        end_time: str = "",
        namespace: typing.List[str] = None,
        engine: str = "",
        datadir: str = "",
        debug: str = '',
        col_width: int = 50,
        max_rows: int = None,
        all_columns: str = None,
        rest_server_ip: str = "",
        rest_server_port: str = "",
        rest_api_key: str = "",
        rest_use_https: str = ""
):
    """set certain contexts for subsequent commands. Cmd is additive"""
    plugin_ctx = context.get_context()
    ctxt = plugin_ctx.ctxt

    if namespace:
        ctxt.namespace = namespace

    if hostname:
        ctxt.hostname = hostname

    if start_time:
        ctxt.start_time = start_time

    if end_time:
        ctxt.end_time = end_time

    if engine:
        plugin_ctx.change_engine(engine)

    if debug:
        ctxt.debug = True

    if datadir:
        if os.path.isdir(datadir):
            ctxt.cfg['data-directory'] = datadir
        else:
            print(f'{datadir} is not a valid directory')

    if col_width:
        ctxt.col_width = int(col_width)

    if max_rows is not None:
        if max_rows == 0:
            ctxt.max_rows = None
        else:
            ctxt.max_rows = max_rows

    if all_columns is not None:
        ctxt.all_columns = all_columns == "yes"

    if pager == 'on':
        ctxt.pager = True
    elif pager == 'off':
        ctxt.pager = False

    if rest_server_ip:
        ctxt.rest_server_ip = rest_server_ip

    if rest_server_port:
        ctxt.rest_server_port = rest_server_port

    if rest_api_key:
        ctxt.rest_api_key = rest_api_key

    if rest_use_https == 'True':
        ctxt.rest_transport = 'https'
    elif rest_use_https == 'False':
        ctxt.rest_transport = 'http'


@command("clear")
@argument("namespace", description="namespace to qualify selection")
@argument("hostname", description="Name of host to qualify selection")
@argument(
    "start_time",
    description="Start of time window in YYYY-MM-dd HH:mm:SS format"
)
@argument("end_time",
          description="End of time window in YYYY-MM-dd HH:mm:SS format")
@argument("pager",
          description="End of time window in YYYY-MM-dd HH:mm:SS format")
@argument(
    "debug", choices=['True'], description="Disable debug mode",
)
def clear_ctxt(
        pager: str = 'off',
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        namespace: str = "",
        debug: str = '',
):
    """clear certain contexts for subsequent commands. Cmd is additive"""
    plugin_ctx = context.get_context()
    ctxt = plugin_ctx.ctxt

    if namespace:
        ctxt.namespace = []

    if hostname:
        ctxt.hostname = []

    if start_time:
        ctxt.start_time = ""

    if end_time:
        ctxt.end_time = ""

    if pager:
        ctxt.pager = False

    if debug:
        ctxt.debug = False


@command('version', help='print the suzieq version')
def sq_version():
    '''Print suzieq version'''
    print_version()


class SqHelpCommand (Command):
    '''Class to handle processing CLI help'''
    HELP = 'Get help on a command'
    thiscmd = ["help", "?"]

    def __init__(self):
        super().__init__()
        self._allcmds = self._sqcmds = None

    def run_interactive(self, cmd, args, raw):
        arglist = args.split()
        return self._help_cmd(*arglist)

    def get_command_names(self):
        return self.thiscmd

    def add_arguments(self, parser):
        parser.add_parser("help")

    def get_help(self, cmd, *args):
        return self.thiscmd[0]

    def get_completions(self, cmd, document, complete_event):
        exploded = document.text.lstrip().split(" ", 1)
        self._build_cmd_verb_list()
        if len(exploded) <= 1:
            if not document.text:
                return [Completion(x)
                        for x in self._sqcmds]
            completions = [Completion(text=x,
                                      start_position=-len(document.text))
                           for x in self._sqcmds
                           if x.startswith(document.text)]
            return completions

        service = exploded[0].lower()
        verb = exploded[1]

        verbs = [x[0].replace('aver', 'assert')
                 for x in self._allcmds[service].metadata.subcommands
                 if service in self._sqcmds]
        if not verb:
            return [Completion(x) for x in verbs]
        completions = [Completion(text=x, start_position=-len(verb))
                       for x in sorted(verbs)
                       if x.startswith(verb)]
        return completions

    def _help_cmd(self, *args):
        """Show help for a service"""
        if len(args) == 0:
            helpcmd = HelpCommand()
            helpcmd.run_interactive('', '', '')
            cprint(
                "Use "
                f"{colored('help <name of command/service> [<verb>]', 'cyan')}"
                " to get more help")
            cprint(f"For example: {colored('help route', 'cyan')}"
                   f" or {colored('help route show', 'cyan')}")
            return
        if len(args) == 1:
            service = args[0]
            verb = ''
        else:
            service = args[0]
            verb = args[1]

        self._build_cmd_verb_list()
        if service in self._sqcmds:
            if verb:
                self._allcmds[service].run_interactive(
                    service, f'help command={verb}', '')
            else:
                self._allcmds[service].run_interactive(service, 'help', '')

    def _build_cmd_verb_list(self):
        if not self._allcmds:
            ctx = context.get_context()
            self._allcmds = ctx.registry.get_all_commands_map()
            self._sqcmds = [x for x in sorted(self._allcmds)
                            if not self._allcmds[x].built_in]



import binascii
import logging
import os
from pathlib import Path
from typing import Callable, Dict, List

import aiofiles
import yaml
from cryptography.fernet import Fernet, InvalidToken
from suzieq.poller.worker.inventory.inventory import Inventory
from suzieq.shared.exceptions import InventorySourceError

logger = logging.getLogger(__name__)


class StaticManagerInventory(Inventory):
    """StaticManager allows to import the inventory from files
    generated by the controller
    """

    def __init__(self, add_task_fn: Callable, **kwargs) -> None:
        worker_id = kwargs.pop('worker-id', '0')

        # Retrieve the key for credential file decryption
        cred_key_str = os.environ.get('SQ_CONTROLLER_POLLER_CRED', None)
        if not cred_key_str:
            raise InventorySourceError(
                'Unable to retrieve the key for decrypting the cred file'
            )
        self._cred_key = cred_key_str.encode('utf-8')

        # Get the paths of both credential and inventory files
        inv_path = os.environ.get('SQ_INVENTORY_PATH', None)
        if not inv_path:
            raise InventorySourceError(
                'Unable to get the inventory path'
            )

        self._inventory_file = Path(inv_path).joinpath(
            f'inv_{worker_id}.yml').resolve()

        self._cred_file = Path(inv_path).joinpath(
            f'cred_{worker_id}').resolve()

        if not self._inventory_file.is_file():
            raise InventorySourceError(
                f'No inventory found at {self._inventory_file}')

        if not self._cred_file.is_file():
            raise InventorySourceError(
                f'No credential file found at {self._cred_file}')

        super().__init__(add_task_fn, **kwargs)

        self._max_outstanding_cmd = int(
            os.environ.get('SQ_MAX_OUTSTANDING_CMD', 0))

    async def _get_device_list(self) -> List[Dict]:

        async with aiofiles.open(str(self._inventory_file), "r") as out_file:
            inventory_data = await out_file.read()
            inventory = yaml.safe_load(inventory_data)

        if not isinstance(inventory, dict):
            raise InventorySourceError('Invalid inventory format. Expected'
                                       f'dict, found {type(inventory)}')

        async with aiofiles.open(str(self._cred_file), "r") as out_file:
            cred_enc_data = await out_file.read()

        # Decrypt the credentials
        try:
            decryptor = Fernet(self._cred_key)
            cred_data = decryptor.decrypt(cred_enc_data.encode('utf-8'))
        except (InvalidToken, ValueError, binascii.Error):
            raise InventorySourceError(
                'The credential decryption key is not valid')
        credentials = yaml.safe_load(cred_data.decode())

        if not isinstance(credentials, dict):
            raise InventorySourceError('Invalid credential file format.')

        # Add credentials to the inventory file
        for k, v in credentials.items():
            if k in inventory:
                inventory[k].update(v)

        return inventory.values()



import re
from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.inventory import InventoryObj


@command("inventory", help="Act on inventory data")
@argument("type", description="Filter by type",
          choices=["fan", "power", "xcvr", "supervisor", "port-adapter",
                   "linecard", "fabric", "midplane", "mx-cb"])
@argument("status", description="Filter by status",
          choices=['present', 'absent'])
@argument("model", description="Filter by model")
@argument("serial", description="Serial number(s), space separated")
@argument("vendor", description="Vendor(s), space separated")
class InventoryCmd(SqTableCommand):
    """Device inventory information such as serial number, cable info etc"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = ' ',
            columns: str = "default",
            type: str = '',     # pylint: disable=redefined-builtin
            status: str = '',
            model: str = '',
            serial: str = '',
            vendor: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=InventoryObj,
        )

        if vendor:
            vendor = re.split(r"\s+(?=[^']*(?:'))", vendor)
        else:
            vendor = []

        if model:
            model = re.split(r"\s+(?=[^']*(?:'))", model)
        else:
            model = []

        self.lvars = {
            'type': type.split(),
            'status': status,
            'model': model,
            'vendor': vendor,
            'serial': serial,
        }



from suzieq.sqobjects.basicobj import SqObject


class EvpnvniObj(SqObject):
    '''The object providing access to the evpnVni table'''

    def __init__(self, **kwargs):
        super().__init__(table='evpnVni', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'vni',
                                'priVtepIp', 'query_str']
        self._valid_assert_args = self._valid_get_args + ['result']
        self._valid_arg_vals = {
            'result': ['all', 'pass', 'fail'],
        }



from typing import Dict, List
from dataclasses import dataclass, field

from suzieq.shared.utils import SUPPORTED_ENGINES


@dataclass
class SqContext:
    '''Context passed between the front-end and back-end'''
    cfg: Dict = field(default_factory=dict)
    schemas: Dict = field(default_factory=dict)

    pager: bool = False
    namespace: str = ''
    hostname: str = ''
    start_time: str = ''
    end_time: str = ''
    exec_time: str = ''
    engine: str = None
    col_width: int = 50
    max_rows: int = 256
    all_columns: bool = False
    debug: bool = False
    sort_fields: List[str] = field(default_factory=list)
    view: str = None
    rest_server_ip: str = '127.0.0.1'
    rest_server_port: int = 8000
    rest_api_key: str = ''
    rest_transport: str = 'https'

    def __post_init__(self):
        # If the engine has not been explicitly set in the context object,
        # get it from the config file
        if not self.engine:
            self.engine = self.cfg.get('ux', {}).get('engine', 'pandas')
            if self.engine == 'rest':
                # See if we can extract the REST info from the REST part
                restcfg = self.cfg.get('rest', {})
                self.rest_server_ip = restcfg.get('address', '127.0.0.1')
                self.rest_server_port = restcfg.get('port', '80')
                if restcfg.get('no-https', False):
                    self.rest_transport = 'http'
                else:
                    self.rest_transport = 'https'
                self.rest_api_key = restcfg.get('API_KEY', '')

        if self.engine not in SUPPORTED_ENGINES:
            raise ValueError(f'Engine {self.engine} not supported')



from dataclasses import dataclass
from pathlib import Path
from importlib.util import find_spec

import streamlit as st

from suzieq.gui.stlit.pagecls import SqGuiPage


@dataclass
class HelpSessionState:
    '''Session state for status page'''
    page: str = 'Help'
    help_on: str = ''


class HelpPage(SqGuiPage):
    '''Page for Path trace page'''
    _title: str = 'Help'
    _state: HelpSessionState = HelpSessionState()
    _helpdir: Path = Path(
        find_spec('suzieq.gui').loader.path).parent / 'help-pages'

    @property
    def add_to_menu(self) -> bool:
        return False

    def build(self):
        self._get_state_from_url()
        self._create_sidebar()
        self._render(None)

    def _create_sidebar(self) -> None:

        filelist = Path(self._helpdir).glob('*.md')
        help_pages = [x.name[:-3] for x in filelist]
        help_on = st.sidebar.selectbox(label='Help On...', options=help_pages)

        return help_on

    def _create_layout(self) -> None:
        pass

    def _render(self, layout: dict) -> None:
        if not self._state.help_on:
            st.info('Nothing to show help about')
            st.stop()

        helptext = ''
        helpfile = self._helpdir / f'{self._state.help_on}.md'
        with open(helpfile, 'r') as f:
            helptext = f.read()

        if helptext:
            st.markdown(helptext)

    def _sync_state(self) -> None:
        pass



from collections import deque
from dataclasses import dataclass, field
from ipaddress import ip_address
from random import randint

import streamlit as st
from pandas.core.frame import DataFrame
from st_aggrid import AgGrid, GridOptionsBuilder, GridUpdateMode
from suzieq.gui.stlit.guiutils import (SuzieqMainPages, gui_get_df,
                                       set_def_aggrid_options)
from suzieq.gui.stlit.pagecls import SqGuiPage
from suzieq.shared.utils import (convert_macaddr_format_to_colon,
                                 validate_macaddr)


@dataclass
class SearchSessionState:
    '''Session state for Search page'''
    page: str = SuzieqMainPages.SEARCH.value
    search_text: str = ''
    past_df = None
    table: str = ''
    namespace: str = ''
    query_str: str = ''
    unique_query: dict = field(default_factory=dict)
    prev_results = deque(maxlen=5)


class SearchPage(SqGuiPage):
    '''Page for Path trace page'''
    _title: str = SuzieqMainPages.SEARCH.value
    _state = SearchSessionState()
    _config_file = st.session_state.get('config_file', '')

    @property
    def add_to_menu(self):
        return True

    def build(self):
        self._get_state_from_url()
        self._create_sidebar()
        layout = self._create_layout()
        self._render(layout)
        self._save_page_url()

    def _create_sidebar(self) -> None:

        state = self._state
        devdf = gui_get_df('device', self._config_file,
                           columns=['namespace', 'hostname'])
        if devdf.empty:
            st.error('Unable to retrieve any namespace info')
            st.stop()

        namespaces = [''] + sorted(devdf.namespace.unique().tolist())
        if not state.namespace:
            nsidx = 0
        else:
            nsidx = namespaces.index(state.namespace)
        namespace = st.sidebar.selectbox('Namespace',
                                         namespaces, key='search_ns',
                                         index=nsidx,
                                         on_change=self._sync_state)

        st.sidebar.markdown(
            """Displays last 5 search results.

You can use search to find specific objects. You can qualify what you're
searching for by qualifying the search term with the type. We support:
- __addresses__: You can qualify a specific table to look for the address.
                 The search string can start with one of the following
                 keywords: __route, mac, arpnd__, to specify which table you
                 want the search to be performed in . If you don't specify a
                 table name, we assume ```network find``` to search for the
                 network attach point for the address. For example,
                 ```arpnd 172.16.1.101``` searches for entries with
                 172.16.1.101 in the IP address column of the arpnd table.
                 Similarly, ```10.0.0.21``` searches for where in the
                 network that IP address is attached to.
- __ASN__: Start the search with the string ```asn``` followed by the ASN
           number. Typing ```asns``` will show you the list of unique ASNs
               across the specified namespaces.
- __VTEP__: Start the search with the string ```vtep``` followed by the VTEP
            IP address. Typing ```vteps``` will show you the list of unique
            VTEPs across the specified namespaces.
- __VNI__: Start the search with the string ```vni``` followed by the VNI
           number.
Typing ```mtus``` will show you the list of unique MTUs across the
specified namespaces.

When specifying a table, you can specify multiple addresses to look for by
 providing the addresses as a space separated values such as
 ```"172.16.1.101 10.0.0.11"``` or
  ```mac "00:01:02:03:04:05 00:21:22:23:24:25"```
  and so on. A combination of mac and IP address can also be specified with
   the address table. Support for more sophisticated search will be added in
    the next few releases.
    """)

        if namespace != state.namespace:
            state.namespace = namespace

    def _create_layout(self) -> dict:
        return {
            'current': st.empty()
        }

    def _render(self, layout) -> None:

        state = self._state
        search_text = st.session_state.search or state.search_text
        query_str, uniq_dict, columns = '', {}, []
        df = DataFrame()
        try:
            query_str, uniq_dict, columns = self._build_query(search_text)
        except ValueError as ve:
            expander = layout['current'].expander(f'Search for {search_text}',
                                                  expanded=True)
            df = DataFrame({'error': [ve]})
            self._draw_aggrid_df(expander, df)

        if state.namespace:
            query_ns = [state.namespace]
        else:
            query_ns = []

        if query_str:
            if state.table == "network":
                df = gui_get_df(state.table,
                                self._config_file,
                                verb='find',
                                namespace=query_ns,
                                view="latest", columns=columns,
                                address=query_str.split())
            else:
                df = gui_get_df(state.table,
                                self._config_file,
                                namespace=query_ns, query_str=query_str,
                                view="latest", columns=columns)
                if not df.empty:
                    df = df.query(query_str) \
                        .drop_duplicates() \
                        .reset_index(drop=True)

            expander = layout['current'].expander(f'Search for {search_text}',
                                                  expanded=True)
            self._draw_aggrid_df(expander, df)

        elif uniq_dict:
            columns = ['namespace'] + uniq_dict['column']
            df = gui_get_df(uniq_dict['table'], self._config_file,
                            namespace=query_ns, view='latest', columns=columns)
            if not df.empty:
                df = df.groupby(by=columns).first().reset_index()

            expander = layout['current'].expander(f'Search for {search_text}',
                                                  expanded=True)
            self._draw_aggrid_df(expander, df)
        elif len(state.prev_results) == 0:
            st.info('Enter a search string to see results, '
                    'see sidebar for examples')

        prev_searches = [search_text]
        for psrch, prev_df in reversed(state.prev_results):
            if psrch in prev_searches:
                continue
            prev_searches.append(psrch)
            expander = st.expander(f'Search for {psrch}', expanded=True)
            self._draw_aggrid_df(expander, prev_df)

        is_error = (not df.empty) and all(df.columns == 'error')
        if ((query_str or uniq_dict) and st.session_state.search and
                (st.session_state.search != state.search_text) and
                not is_error):
            state.prev_results.append((search_text, df))
            state.search_text = st.session_state.search

    def _draw_aggrid_df(self, expander, df):

        with expander:
            if df.empty:
                st.info('No matching result found')
            elif 'error' in df:
                st.error(df['error'][0])
            else:

                gb = GridOptionsBuilder.from_dataframe(df)
                gb.configure_pagination(paginationPageSize=25)

                gb.configure_default_column(floatingFilter=True,
                                            editable=False,
                                            selectable=False)

                gb.configure_grid_options(
                    domLayout='normal', preventDefaultOnContextMenu=True)
                gridOptions = set_def_aggrid_options(gb.build())

                if df.shape[0] == 1:
                    height = 150
                elif df.shape[0] < 4:
                    height = 200
                else:
                    height = 400
                _ = AgGrid(
                    df,
                    height=height,
                    gridOptions=gridOptions,
                    allow_unsafe_jscode=True,
                    update_mode=GridUpdateMode.NO_UPDATE,
                    theme='streamlit',
                    key=str(randint(1, 10000000))
                )

    def _sync_state(self) -> None:
        pass

    # pylint: disable=too-many-statements
    def _build_query(self, search_text: str):
        '''Build the appropriate query for the search'''

        state = self._state
        search_text = search_text.replace('\"', '').replace('\'', '')
        if not search_text:
            return '', {}, []

        unique_query = {}

        addrs = search_text.split()
        if not addrs:
            return '', {}, []

        query_str = disjunction = ''
        columns = ['default']

        if addrs[0] not in ['mac', 'macs', 'route', 'routes',
                            'arpnd', 'address', 'vtep', 'vteps',
                            'asn', 'asns', 'vlan', 'vlans',
                            'mtu', 'mtus']:
            try:
                ip_address(addrs[0])
            except ValueError:
                if not validate_macaddr(addrs[0]):
                    raise ValueError('Invalid keyword or IP/Mac address '
                                     f'"{addrs[0]}"')

        if addrs[0].startswith('mac'):
            state.table = 'macs'
            addrs = addrs[1:]
        elif addrs[0].startswith('route'):
            state.table = 'routes'
            addrs = addrs[1:]
        elif addrs[0] == 'arpnd':
            state.table = 'arpnd'
            addrs = addrs[1:]
        elif addrs[0].startswith('address'):
            state.table = 'network'
            search_text = ' '.join(addrs[1:])
        elif addrs[0].startswith('vtep'):
            state.table = 'evpnVni'
            if addrs[0] != 'vteps':
                query_str = (f'priVtepIp.isin({addrs[1:]}) or '
                             f'secVtepIp.isin({addrs[1:]})')
                columns = ['namespace', 'hostname', 'priVtepIp',
                           'secVtepIp']
        elif addrs[0].startswith('vni'):
            state.table = 'evpnVni'
            if addrs[0] != 'vnis':
                try:
                    vnis = [int(x) for x in addrs[1:]]
                except ValueError:
                    vnis = []
                query_str = f'vni.isin({vnis})'
                columns = ['namespace', 'hostname', 'vni']
        elif addrs[0].startswith('asn'):
            state.table = 'bgp'
            if addrs[0] != "asns":
                try:
                    asns = [int(x) for x in addrs[1:]]
                except ValueError:
                    asns = []
                query_str = f'asn.isin({asns})'
                columns = ['namespace', 'hostname', 'asn']
        elif addrs[0].startswith('vlan'):
            state.table = 'vlan'
            if addrs[0] != "vlans":
                try:
                    vlans = [int(x) for x in addrs[1:]]
                except ValueError:
                    vlans = []
                query_str = f'vlan.isin({vlans})'
                columns = ['namespace', 'hostname', 'vlan']
        elif addrs[0].startswith('mtu'):
            state.table = 'interface'
            if addrs[0] != "mtus":
                try:
                    mtus = [int(x) for x in addrs[1:]]
                except ValueError:
                    mtus = []
                query_str = f'mtu.isin({mtus})'
                columns = ['namespace', 'hostname', 'mtu']
        else:
            state.table = 'network'

        if state.table == 'network':
            return search_text, unique_query, columns

        for addr in addrs:
            if addr.lower() == 'vteps':
                unique_query = {'table': 'evpnVni',
                                'column': ['priVtepIp', 'secVtepIp'],
                                'colname': 'vteps'}
            elif addr.lower() == 'vnis':
                unique_query = {'table': 'evpnVni',
                                'column': ['vni'], 'colname': 'vnis'}
            elif addr.lower() == 'asns':
                unique_query = {'table': 'bgp', 'column': ['asn', 'peerAsn'],
                                'colname': 'asns'}
            elif addr.lower() == 'vlans':
                unique_query = {'table': 'vlan', 'column': ['vlan'],
                                'colname': 'vlans'}
            elif addr.lower() == 'mtus':
                unique_query = {'table': 'interfaces', 'column': ['mtu'],
                                'colname': 'mtus'}

            elif '::' in addr:
                if state.table == 'arpnd':
                    query_str += f' {disjunction} ipAddress == "{addr}" '
                elif state.table == 'routes':
                    query_str += f'{disjunction} prefix == "{addr}" '
                else:
                    query_str += f' {disjunction} ' \
                        f'ip6AddressList.str.startswith("{addr}/") '
            elif ':' in addr and state.table in ['macs', 'arpnd']:
                query_str += f' {disjunction} macaddr == "{addr}" '
            elif state.table in ['macs', 'arpnd', 'routes']:
                try:
                    addr = ip_address(addr)
                    macaddr = None
                except ValueError:
                    macaddr = convert_macaddr_format_to_colon(addr)
                    addr = None

                if state.table == "macs":
                    query_str = f'{disjunction} macaddr == "{macaddr}" '
                elif state.table == 'arpnd':
                    if addr:
                        query_str += f' {disjunction} ipAddress == "{addr}" '
                    elif macaddr:
                        query_str += f' {disjunction} macaddr == "{macaddr}" '
                elif state.table == 'routes':
                    query_str += f'{disjunction} prefix == "{addr}" '
                else:
                    query_str = ''

            if not disjunction:
                disjunction = 'or'

        state.query_str = query_str
        state.unique_query = unique_query
        return query_str, unique_query, columns



from typing import List
from suzieq.gui.stlit.pagecls import SqGuiPage


def get_menu_pages() -> List[str]:
    '''Return a list of page names to be included in the menu

    Returns:
        List[str]: sorted list of all page names
    '''
    pages = SqGuiPage.get_plugins()
    return sorted([key for key, obj in pages.items() if obj.add_to_menu])


def get_page(page_name: str):
    '''Return the object associated with the page requested

    :param page_name: str, page name for which object is requested
    :returns:
    :rtype: The child class of SqGuiPage that corresponds to the page
            or raises ModuleNotFoundError if the table is not found

    '''

    pages = SqGuiPage.get_plugins()
    if page_name not in pages:
        raise ModuleNotFoundError(f'Page {page_name} not found')

    return pages[page_name]


__all__ = ['get_menu_pages', 'get_page']
